#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <GL/glut.h>

#include "vector.h"
#include "Obj.h"
#include "Block.h"
#include "Saucer.h"

struct _Saucer {
  vector_t position;
  vector_t angle;

  int count;
  vector_t v;
  int count_angle;
  vector_t angle_v;

  int track_num;
  int sector_num;

  ObjList blocks;
};

Saucer Saucer_Create(double x, double y, double z,
		     int track_num, int sector_num)
{
  Saucer saucer;
  Block block;
  ObjList sectors;
  int t, s;

  saucer = malloc(sizeof(*saucer));
  memset(saucer, 0, sizeof(*saucer));

  vector_set(&saucer->position, x, y, z);
  vector_zero(&saucer->angle);

  saucer->count = 0;
  vector_zero(&saucer->v);

  saucer->count_angle = 0;
  vector_zero(&saucer->angle_v);

  saucer->track_num  = track_num;
  saucer->sector_num = sector_num;

  saucer->blocks = ObjList_Create();

  for (t = 0; t < track_num; t++) {
    sectors = ObjList_Create();
    for (s = 0; s < sector_num; s++) {
      block = Block_Create(t, s);
      ObjList_InsertObjToEnd(sectors, block, (ObjDestructor)Block_Destroy);
    }
    ObjList_InsertObjToEnd(saucer->blocks, sectors,
			   (ObjDestructor)ObjList_Destroy);
  }

  return saucer;
}

int Saucer_Proc(Saucer saucer)
{
  Block block;
  ObjList sectors;
  ObjListData obj, obj2;

  if (saucer->count > 0) {
    saucer->count--;
    vector_add(&saucer->position, &saucer->position, &saucer->v);
  }
  if (saucer->count_angle > 0) {
    saucer->count_angle--;
    vector_add(&saucer->angle, &saucer->angle, &saucer->angle_v);
  }

  for (obj = ObjList_GetStart(saucer->blocks);
       !ObjList_IsEndEdge(saucer->blocks, obj);
       obj = ObjListData_GetNext(obj)) {
    sectors = (ObjList)ObjListData_GetObj(obj);
    for (obj2 = ObjList_GetStart(sectors);
	 !ObjList_IsEndEdge(sectors, obj2);
	 obj2 = ObjListData_GetNext(obj2)) {
      block = (Block)ObjListData_GetObj(obj2);
      Block_Proc(block);
    }
  }

  return 1;
}

int Saucer_MakePolygon(Saucer saucer)
{
  Block block;
  ObjList sectors;
  ObjListData obj, obj2;

  glTranslated(saucer->position.x, saucer->position.y, saucer->position.z);
  glRotated(saucer->angle.x, 1.0, 0.0, 0.0);
  glRotated(saucer->angle.y, 0.0, 1.0, 0.0);
  glRotated(saucer->angle.z, 0.0, 0.0, 1.0);
  glScaled(0.5, 0.5, 0.5);

  for (obj = ObjList_GetStart(saucer->blocks);
       !ObjList_IsEndEdge(saucer->blocks, obj);
       obj = ObjListData_GetNext(obj)) {
    sectors = (ObjList)ObjListData_GetObj(obj);
    for (obj2 = ObjList_GetStart(sectors);
	 !ObjList_IsEndEdge(sectors, obj2);
	 obj2 = ObjListData_GetNext(obj2)) {
      block = (Block)ObjListData_GetObj(obj2);
      glPushMatrix();
      Block_MakePolygon(block, saucer->track_num, saucer->sector_num);
      glPopMatrix();
    }
  }

  return 0;
}

Block Saucer_GetBlock(Saucer saucer, int track, int sector)
{
  Block block;
  ObjList sectors;
  ObjListData obj, obj2;

  for (obj = ObjList_GetStart(saucer->blocks);
       !ObjList_IsEndEdge(saucer->blocks, obj);
       obj = ObjListData_GetNext(obj)) {

    if (track > 0) {
      track--;
      continue;
    }

    sectors = (ObjList)ObjListData_GetObj(obj);
    for (obj2 = ObjList_GetStart(sectors);
	 !ObjList_IsEndEdge(sectors, obj2);
	 obj2 = ObjListData_GetNext(obj2)) {
      if (sector > 0) {
	sector--;
	continue;
      }
      block = (Block)ObjListData_GetObj(obj2);
      return block;
    }
  }

  return NULL;
}

vector_t *Saucer_GetPosition(Saucer saucer)
{
  return &saucer->position;
}

vector_t *Saucer_GetAngle(Saucer saucer)
{
  return &saucer->angle;
}

int Saucer_GetTrackNum(Saucer saucer)
{
  return saucer->track_num;
}

int Saucer_GetSectorNum(Saucer saucer)
{
  return saucer->sector_num;
}

int Saucer_AddTrackNum(Saucer saucer, int n)
{
  Block block;
  ObjList sectors;
  int t, s;

  for (t = saucer->track_num; t < saucer->track_num + n; t++) {
    sectors = ObjList_Create();
    for (s = 0; s < saucer->sector_num; s++) {
      block = Block_Create(t, s);
      ObjList_InsertObjToEnd(sectors, block, (ObjDestructor)Block_Destroy);
    }
    ObjList_InsertObjToEnd(saucer->blocks, sectors,
			   (ObjDestructor)ObjList_Destroy);
  }
  saucer->track_num += n;

  return saucer->track_num;
}

int Saucer_AddSectorNum(Saucer saucer, int n)
{
  Block block;
  ObjList sectors;
  ObjListData obj;
  int t, s;

  t = 0;
  for (obj = ObjList_GetStart(saucer->blocks);
       !ObjList_IsEndEdge(saucer->blocks, obj);
       obj = ObjListData_GetNext(obj)) {
    sectors = (ObjList)ObjListData_GetObj(obj);
    for (s = saucer->sector_num; s < saucer->sector_num + n; s++) {
      block = Block_Create(t, s);
      ObjList_InsertObjToEnd(sectors, block, (ObjDestructor)Block_Destroy);
    }
    t++;
  }
  saucer->sector_num += n;

  return saucer->sector_num;
}

int Saucer_Move(Saucer saucer, int count, double x, double y, double z)
{
  saucer->count = count;
  vector_set(&saucer->v, x, y, z);
  vector_sub(&saucer->v, &saucer->v, &saucer->position);
  vector_div(&saucer->v, &saucer->v, count);
  return 0;
}

int Saucer_Rotate(Saucer saucer, int count, double x, double y, double z)
{
  saucer->count_angle = count;
  vector_set(&saucer->angle_v, x, y, z);
  vector_normalize_angle_positive(&saucer->angle_v);
  vector_normalize_angle_positive(&saucer->angle);
  vector_sub(&saucer->angle_v, &saucer->angle_v, &saucer->angle);
  vector_normalize_angle_sym(&saucer->angle_v);
  vector_div(&saucer->angle_v, &saucer->angle_v, count);
  return 0;
}
