#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "vector.h"
#include "pattern.h"

static vector_t updir   = { 0.0, 0.0,  1.0 };
static vector_t downdir = { 0.0, 0.0, -1.0 };
#define DELTA 0.0001

static int clear_points(int n, vector_t *points)
{
  int i;
  for (i = 0; i < n; i++)
    vector_zero(&points[i]);
  return 0;
}

static int clear_angles(int n, vector_t *angles)
{
  int i;
  for (i = 0; i < n; i++)
    vector_zero(&angles[i]);
  return 0;
}

static int clear(int n, vector_t *points, vector_t *angles)
{
  clear_points(n, points);
  clear_angles(n, angles);
  return 0;
}

static int pattern_default(int n, vector_t *points, vector_t *angles)
{
  int i;
  clear(n, points, angles);
  for (i = 0; i < n; i++) {
    points[i].x = cos(2.0 * M_PI * i * 0.3) * 0.8;
    points[i].y = sin(2.0 * M_PI * i * 0.3) * 0.8;
    points[i].z = (double)(i + 1) / n * 2.0 - 1.0;
  }
  return 0;
}

static int pattern_random(int n, vector_t *points, vector_t *angles)
{
  int i;
  clear(n, points, angles);
  for (i = 0; i < n; i++) {
    points[i].x = ((rand() % 201) - 100) * 0.01;
    points[i].y = ((rand() % 201) - 100) * 0.01;
    points[i].z = ((rand() % 201) - 100) * 0.01;
  }
  return 0;
}

static int pattern_random_angle(int n, vector_t *points, vector_t *angles)
{
  int i;
  clear(n, points, angles);
  for (i = 0; i < n; i++) {
    points[i].x = ((rand() % 201) - 100) * 0.01;
    points[i].y = ((rand() % 201) - 100) * 0.01;
    points[i].z = ((rand() % 201) - 100) * 0.01;
    vector_set(&angles[i], rand() % 360, rand() % 360, rand() % 360);
  }
  return 0;
}

static int pattern_circle(int n, vector_t *points, vector_t *angles)
{
  int i;
  vector_t dir;
  clear(n, points, angles);
  for (i = 0; i < n; i++) {
    points[i].x = cos(2.0 * M_PI * ((double)i / n) + DELTA) * 1.5;
    points[i].y = sin(2.0 * M_PI * ((double)i / n) + DELTA) * 1.5;
    vector_mul(&dir, &points[i], -1.0);
    vector_calc_angle(&angles[i], &updir, &dir);
  }
  return 0;
}

static int pattern_circle_alt(int n, vector_t *points, vector_t *angles)
{
  int i;
  vector_t dir;
  clear(n, points, angles);
  for (i = 0; i < n; i++) {
    points[i].x = cos(2.0 * M_PI * ((double)i / n) + DELTA) * 1.5;
    points[i].y = sin(2.0 * M_PI * ((double)i / n) + DELTA) * 1.5;
    points[i].z = ((i % 2) - 0.5) * 0.2;
    vector_mul(&dir, &points[i], -1.0);
    vector_calc_angle(&angles[i], &updir, &dir);
  }
  return 0;
}

static int pattern_circle_alt_rev(int n, vector_t *points, vector_t *angles)
{
  int i;
  vector_t dir;
  clear(n, points, angles);
  for (i = 0; i < n; i++) {
    points[i].x = cos(2.0 * M_PI * ((double)i / n) + DELTA) * 1.5;
    points[i].y = sin(2.0 * M_PI * ((double)i / n) + DELTA) * 1.5;
    points[i].z = ((i % 2) - 0.5) * 0.2;
    vector_mul(&dir, &points[i], -1.0);
    vector_calc_angle(&angles[i], (i % 2) ? &updir : &downdir, &dir);
  }
  return 0;
}

static int pattern_circle_alt_out(int n, vector_t *points, vector_t *angles)
{
  int i;
  clear(n, points, angles);
  for (i = 0; i < n; i++) {
    points[i].x = cos(2.0 * M_PI * ((double)i / n) + DELTA) * 1.5;
    points[i].y = sin(2.0 * M_PI * ((double)i / n) + DELTA) * 1.5;
    points[i].z = 0.0;
    vector_calc_angle(&angles[i], &points[i], (i % 2) ? &updir : &downdir);
    points[i].z = ((i % 2) - 0.5) * 0.7;
  }
  return 0;
}

static int pattern_circle_alt_out2(int n, vector_t *points, vector_t *angles)
{
  int i;
  clear(n, points, angles);
  for (i = 0; i < n; i++) {
    points[i].x = cos(2.0 * M_PI * ((double)i / n) + DELTA) * 1.5;
    points[i].y = sin(2.0 * M_PI * ((double)i / n) + DELTA) * 1.5;
    points[i].z = ((i % 2) - 0.5) * 0.7;
    vector_calc_angle(&angles[i], &points[i], (i % 2) ? &updir : &downdir);
  }
  return 0;
}

static int pattern_screw(int n, vector_t *points, vector_t *angles)
{
  int i;
  vector_t dir;
  clear(n, points, angles);
  for (i = 0; i < n; i++) {
    points[i].x = cos(2.0 * M_PI * ((double)i / n) + DELTA) * 1.2;
    points[i].y = sin(2.0 * M_PI * ((double)i / n) + DELTA) * 1.2;
    points[i].z = (double)i / n * 2.0 - 1.0;
    vector_mul(&dir, &points[i], -1.0);
    vector_calc_angle(&angles[i], &updir, &dir);
  }
  return 0;
}

static int pattern_screw_many(int n, vector_t *points, vector_t *angles)
{
  int i;
  vector_t dir;
  clear(n, points, angles);
  for (i = 0; i < n; i++) {
    points[i].x = cos(2.0 * M_PI * ((double)(i * 2.5) / n) + DELTA) * 1.2;
    points[i].y = sin(2.0 * M_PI * ((double)(i * 2.5) / n) + DELTA) * 1.2;
    points[i].z = (double)i / n * 2.0 - 1.0;
    vector_mul(&dir, &points[i], -1.0);
    vector_calc_angle(&angles[i], &updir, &dir);
  }
  return 0;
}

static int pattern_screw_many_out(int n, vector_t *points, vector_t *angles)
{
  int i;
  clear(n, points, angles);
  for (i = 0; i < n; i++) {
    points[i].x = cos(2.0 * M_PI * ((double)(i * 2.5) / n) + DELTA) * 1.2;
    points[i].y = sin(2.0 * M_PI * ((double)(i * 2.5) / n) + DELTA) * 1.2;
    points[i].z = (double)i / n * 2.0 - 1.0;
    vector_calc_angle(&angles[i], &points[i], &updir);
  }
  return 0;
}

static int pattern_2circle(int n, vector_t *points, vector_t *angles)
{
  int i;
  vector_t dir;
  clear(n, points, angles);
  for (i = 0; i < n; i++) {
    points[i].x = cos(2.0 * M_PI * ((double)i / n) + DELTA) * 1.2;
    points[i].y = sin(2.0 * M_PI * ((double)i / n) + DELTA) * 1.2;
    points[i].z = ((i % 2) * 2 - 1) * 0.4;
    vector_mul(&dir, &points[i], -1.0);
    vector_calc_angle(&angles[i], &updir, &dir);
  }
  return 0;
}

static int pattern_2circle_alt(int n, vector_t *points, vector_t *angles)
{
  int i;
  vector_t dir;
  clear(n, points, angles);
  for (i = 0; i < n; i++) {
    points[i].x = cos(2.0 * M_PI * ((double)i / n) + DELTA) * 1.2;
    points[i].y = sin(2.0 * M_PI * ((double)i / n) + DELTA) * 1.2;
    points[i].z = ((i % 2) * 2 - 1) * 0.4 + ((i / 2) % 2) * 0.2;
    vector_mul(&dir, &points[i], -1.0);
    vector_calc_angle(&angles[i], &updir, &dir);
  }
  return 0;
}

static int pattern_2circle_alt_out(int n, vector_t *points, vector_t *angles)
{
  int i;
  clear(n, points, angles);
  for (i = 0; i < n; i++) {
    points[i].x = cos(2.0 * M_PI * ((double)i / n) + DELTA) * 1.2;
    points[i].y = sin(2.0 * M_PI * ((double)i / n) + DELTA) * 1.2;
    points[i].z = ((i % 2) * 2 - 1) * 0.4 + ((i / 2) % 2) * 0.2;
    vector_calc_angle(&angles[i], &points[i], (i % 2) ? &updir : &downdir);
  }
  return 0;
}

static int pattern_bridge(int n, vector_t *points, vector_t *angles)
{
  int i;
  vector_t dir;
  clear(n, points, angles);
  for (i = 0; i < n; i++) {
    points[i].x = cos(2.0 * M_PI * ((double)(i / 4) / (n / 4)) + DELTA) * 1.2;
    points[i].y = sin(2.0 * M_PI * ((double)(i / 4) / (n / 4)) + DELTA) * 1.2;
    points[i].z = ((i % 4) - 1.5) * 0.5;
    vector_mul(&dir, &points[i], -1.0);
    vector_calc_angle(&angles[i], &updir, &dir);
  }
  return 0;
}

static int pattern_bridge_alt(int n, vector_t *points, vector_t *angles)
{
  int i;
  vector_t dir;
  clear(n, points, angles);
  for (i = 0; i < n; i++) {
    points[i].x = cos(2.0 * M_PI * (((double)(i / 4) + 0.5 * (i % 2)) / (n / 4)) + DELTA) * 1.2;
    points[i].y = sin(2.0 * M_PI * (((double)(i / 4) + 0.5 * (i % 2)) / (n / 4)) + DELTA) * 1.2;
    points[i].z = ((i % 4) - 1.5) * 0.5;
    vector_mul(&dir, &points[i], -1.0);
    vector_calc_angle(&angles[i], &updir, &dir);
  }
  return 0;
}

static int pattern_bridge_alt_out(int n, vector_t *points, vector_t *angles)
{
  int i;
  clear(n, points, angles);
  for (i = 0; i < n; i++) {
    points[i].x = cos(2.0 * M_PI * (((double)(i / 4) + 0.5 * (i % 2)) / (n / 4)) + DELTA) * 1.2;
    points[i].y = sin(2.0 * M_PI * (((double)(i / 4) + 0.5 * (i % 2)) / (n / 4)) + DELTA) * 1.2;
    points[i].z = 0.0;
    vector_calc_angle(&angles[i], &points[i], &updir);
    points[i].z = ((i % 4) - 1.5) * 0.5;
  }
  return 0;
}

static int pattern_bridge_alt_out2(int n, vector_t *points, vector_t *angles)
{
  int i;
  clear(n, points, angles);
  for (i = 0; i < n; i++) {
    points[i].x = cos(2.0 * M_PI * (((double)(i / 4) + 0.5 * (i % 2)) / (n / 4)) + DELTA) * 1.2;
    points[i].y = sin(2.0 * M_PI * (((double)(i / 4) + 0.5 * (i % 2)) / (n / 4)) + DELTA) * 1.2;
    points[i].z = ((i % 4) - 1.5) * 0.5;
    vector_calc_angle(&angles[i], &points[i], &updir);
  }
  return 0;
}

static int pattern_tower(int n, vector_t *points, vector_t *angles)
{
  int i;
  clear(n, points, angles);
  for (i = 0; i < n; i++) {
    points[i].z = (double)i / n * 2.0 - 1.0;
  }
  return 0;
}

static int pattern_flat(int n, vector_t *points, vector_t *angles)
{
  int i;
  clear(n, points, angles);
  for (i = 0; i < n; i++) {
    if (i < 3) {
      points[i].x = cos(2.0 * M_PI * ((double)i / 3)) * 0.5;
      points[i].y = sin(2.0 * M_PI * ((double)i / 3)) * 0.5;
    } else if (i < 9) {
      points[i].x = cos(2.0 * M_PI * ((double)(i - 3) / 6)) * 1.2;
      points[i].y = sin(2.0 * M_PI * ((double)(i - 3) / 6)) * 1.2;
    } else {
      points[i].x = cos(2.0 * M_PI * ((double)(i - 9) / (n - 9))) * 1.9;
      points[i].y = sin(2.0 * M_PI * ((double)(i - 9) / (n - 9))) * 1.9;
    }
  }
  return 0;
}

static int pattern_flat_alt(int n, vector_t *points, vector_t *angles)
{
  int i;
  clear(n, points, angles);
  for (i = 0; i < n; i++) {
    if (i < 3) {
      points[i].x = cos(2.0 * M_PI * ((double)i / 3)) * 0.5;
      points[i].y = sin(2.0 * M_PI * ((double)i / 3)) * 0.5;
    } else if (i < 9) {
      points[i].x = cos(2.0 * M_PI * ((double)(i - 3) / 6)) * 1.2;
      points[i].y = sin(2.0 * M_PI * ((double)(i - 3) / 6)) * 1.2;
    } else {
      points[i].x = cos(2.0 * M_PI * ((double)(i - 9) / (n - 9))) * 1.9;
      points[i].y = sin(2.0 * M_PI * ((double)(i - 9) / (n - 9))) * 1.9;
    }
    points[i].z = (i % 2) * 0.2;
  }
  return 0;
}

static int pattern_flat_random(int n, vector_t *points, vector_t *angles)
{
  int i;
  clear(n, points, angles);
  for (i = 0; i < n; i++) {
    if (i < 3) {
      points[i].x = cos(2.0 * M_PI * ((double)i / 3)) * 0.5;
      points[i].y = sin(2.0 * M_PI * ((double)i / 3)) * 0.5;
    } else if (i < 9) {
      points[i].x = cos(2.0 * M_PI * ((double)(i - 3) / 6)) * 1.2;
      points[i].y = sin(2.0 * M_PI * ((double)(i - 3) / 6)) * 1.2;
    } else {
      points[i].x = cos(2.0 * M_PI * ((double)(i - 9) / (n - 9))) * 1.9;
      points[i].y = sin(2.0 * M_PI * ((double)(i - 9) / (n - 9))) * 1.9;
    }
    points[i].z = ((rand() % 5) - 2) * 0.1;
  }
  return 0;
}

static int pattern_flat_mount(int n, vector_t *points, vector_t *angles)
{
  int i;
  clear(n, points, angles);
  for (i = 0; i < n; i++) {
    if (i < 3) {
      points[i].x = cos(2.0 * M_PI * ((double)i / 3)) * 0.7;
      points[i].y = sin(2.0 * M_PI * ((double)i / 3)) * 0.7;
      points[i].z = 0.7;
    } else if (i < 9) {
      points[i].x = cos(2.0 * M_PI * ((double)(i - 3) / 6)) * 1.0;
      points[i].y = sin(2.0 * M_PI * ((double)(i - 3) / 6)) * 1.0;
      points[i].z = 0.2;
    } else {
      points[i].x = cos(2.0 * M_PI * ((double)(i - 9) / (n - 9))) * 1.5;
      points[i].y = sin(2.0 * M_PI * ((double)(i - 9) / (n - 9))) * 1.5;
      points[i].z = -0.3;
    }
  }
  return 0;
}

static int pattern_flat_mount_rev(int n, vector_t *points, vector_t *angles)
{
  int i;
  clear(n, points, angles);
  for (i = 0; i < n; i++) {
    if (i < 3) {
      points[i].x = cos(2.0 * M_PI * ((double)i / 3)) * 0.7;
      points[i].y = sin(2.0 * M_PI * ((double)i / 3)) * 0.7;
      points[i].z = -0.5;
    } else if (i < 9) {
      points[i].x = cos(2.0 * M_PI * ((double)(i - 3) / 6)) * 1.0;
      points[i].y = sin(2.0 * M_PI * ((double)(i - 3) / 6)) * 1.0;
      points[i].z = 0.0;
    } else {
      points[i].x = cos(2.0 * M_PI * ((double)(i - 9) / (n - 9))) * 1.5;
      points[i].y = sin(2.0 * M_PI * ((double)(i - 9) / (n - 9))) * 1.5;
      points[i].z = 0.5;
    }
  }
  return 0;
}

static int pattern_square(int n, vector_t *points, vector_t *angles)
{
  int i;
  clear(n, points, angles);
  for (i = 0; i < n; i++) {
    points[i].x = ((i % 4) - 1.5) * 0.8;
    points[i].z = ((i / 4) - (((n + 3) / 4) - 1) * 0.5) * 0.4;
  }
  return 0;
}

static int pattern_square_alt(int n, vector_t *points, vector_t *angles)
{
  int i;
  clear(n, points, angles);
  for (i = 0; i < n; i++) {
    points[i].x = ((i % 4) - 1.5) * 0.8;
    points[i].z = ((i / 4) - (((n + 3) / 4) - 1) * 0.5) * 0.4 + (i % 2) * 0.1;
  }
  return 0;
}

static int pattern_circle_virt(int n, vector_t *points, vector_t *angles)
{
  int i;
  clear(n, points, angles);
  for (i = 0; i < n; i++) {
    points[i].x = cos(2.0 * M_PI * ((double)i / n) + DELTA) * 1.2;
    points[i].z = sin(2.0 * M_PI * ((double)i / n) + DELTA) * 1.2;
  }
  return 0;
}

static int pattern_circle_virt_double(int n, vector_t *points, vector_t *angles)
{
  int i;
  clear(n, points, angles);
  for (i = 0; i < n; i++) {
    if ((i % 2) == 0) {
      points[i].x = cos(2.0 * M_PI * ((double)(i + 1.0) / n) + DELTA) * 1.2;
      points[i].z = sin(2.0 * M_PI * ((double)(i + 1.0) / n) + DELTA) * 1.2;
    } else {
      points[i].y = cos(2.0 * M_PI * ((double)(i + 0.0) / n) + DELTA) * 1.2;
      points[i].z = sin(2.0 * M_PI * ((double)(i + 0.0) / n) + DELTA) * 1.2;
    }
  }
  return 0;
}

static int pattern_circle_virt_double_out(int n, vector_t *points, vector_t *angles)
{
  int i;
  clear(n, points, angles);
  for (i = 0; i < n; i++) {
    if ((i % 2) == 0) {
      points[i].x = cos(2.0 * M_PI * ((double)(i + 1.0) / n) + DELTA) * 1.2;
      points[i].z = sin(2.0 * M_PI * ((double)(i + 1.0) / n) + DELTA) * 1.2;
    } else {
      points[i].y = cos(2.0 * M_PI * ((double)(i + 0.0) / n) + DELTA) * 1.2;
      points[i].z = sin(2.0 * M_PI * ((double)(i + 0.0) / n) + DELTA) * 1.2;
    }
    vector_calc_angle(&angles[i], &points[i], &updir);
  }
  return 0;
}

static int pattern_ball_random(int n, vector_t *points, vector_t *angles)
{
  int i, j, num;
  vector_t vd, v;
  double r;

  vector_set(&points[0], 0.0, 0.0, 1.0);
  clear_angles(n, angles);

  num = 128;

  while (1) {
    for (i = 1; i < n; i++) {
      vector_zero(&v);
      for (j = 0; j < n; j++) {
	if (i == j)
	  continue;
	vector_sub(&vd, &points[i], &points[j]);
	while ((r = vector_length(&vd)) == 0.0)
	  vector_rand(&vd);
	vector_normalize(&vd, 1.0 / (r*r));
	vector_add(&vd, &vd, &points[i]);
	vector_normalize(&vd, 1.0);
	vector_sub(&vd, &vd, &points[i]);
	vector_add(&v, &v, &vd);
      }
      if (vector_length(&v) > 0.5)
	vector_normalize(&v, 0.5);
      vector_add(&points[i], &points[i], &v);
      vector_normalize(&points[i], 1.0);
      num--;
    }
    if (num < 0)
      break;
  }

  return 0;
}

static int pattern_ball_random_out(int n, vector_t *points, vector_t *angles)
{
  int i;
  pattern_ball_random(n, points, angles);
  for (i = 0; i < n; i++)
    vector_calc_angle(&angles[i], &points[i], &updir);
  return 0;
}

typedef int (*pattern_func_t)(int n, vector_t *points, vector_t *angles);
static pattern_func_t funcs[] = {
  pattern_default,
  pattern_random, pattern_random_angle,
  pattern_circle, pattern_circle_alt, pattern_circle_alt_rev,
  pattern_circle_alt_out, pattern_circle_alt_out2,
  pattern_screw, pattern_screw_many, pattern_screw_many_out,
  pattern_2circle, pattern_2circle_alt, pattern_2circle_alt_out,
  pattern_bridge, pattern_bridge_alt,
  pattern_bridge_alt_out, pattern_bridge_alt_out2,
  pattern_tower,
  pattern_flat, pattern_flat_alt, pattern_flat_random,
  pattern_flat_mount, pattern_flat_mount_rev,
  pattern_square, pattern_square_alt,
  pattern_circle_virt, pattern_circle_virt_double,
  pattern_circle_virt_double_out,
  pattern_ball_random, pattern_ball_random_out,
};

int pattern_set(int pattern, int n, vector_t *points, vector_t *angles)
{
  pattern_func_t f;

  f = funcs[pattern % (sizeof(funcs) / sizeof(funcs[0]))];
  f(n, points, angles);

  return 0;
}
